@@ -7,6 +7,11 @@ from sys import version_info
 import logging
 import re
 
+if version_info.major == 2:
+    from urllib2 import unquote
+else:
+    from urllib.parse import unquote
+
 GCM_URL = 'https://gcm-http.googleapis.com/gcm/send'
 
 
@@ -77,7 +82,7 @@ def group_response(response, registration_ids, key):
     # Filter by key
     filtered = ((reg_id, res[key]) for reg_id, res in mapping if key in res)
     # Grouping of errors and mapping of ids
-    if key in ['registration_id','message_id']:
+    if key in ['registration_id', 'message_id']:
         grouping = dict(filtered)
     else:
         grouping = defaultdict(list)
@@ -133,29 +138,29 @@ class Payload(object):
     def validate_time_to_live(self, value):
         if not (0 <= value <= self.GCM_TTL):
             raise GCMInvalidTtlException("Invalid time to live value")
-    
+
     def validate_registration_ids(self, registration_ids):
 
         if len(registration_ids) > 1000:
             raise GCMTooManyRegIdsException("Exceded number of registration_ids")
-    
+
     def validate_to(self, value):
         if not re.match(Payload.topicPattern, value):
-            raise GCMInvalidInputException("Invalid topic name: {0}! Does not match the {1} pattern".format(value, Payload.topicPattern))
-    
+            raise GCMInvalidInputException(
+                "Invalid topic name: {0}! Does not match the {1} pattern".format(value, Payload.topicPattern))
+
     @property
     def body(self):
         raise NotImplementedError
 
 
 class PlaintextPayload(Payload):
-
     @property
     def body(self):
         # Safeguard for backwards compatibility
         if 'registration_id' not in self.__dict__:
             self.__dict__['registration_id'] = self.__dict__.pop(
-                'registration_ids', None
+                    'registration_ids', None
             )
         # Inline data for for plaintext request
         data = self.__dict__.pop('data')
@@ -165,18 +170,17 @@ class PlaintextPayload(Payload):
 
 
 class JsonPayload(Payload):
-
     @property
     def body(self):
         return json.dumps(self.__dict__)
 
 
 class GCM(object):
-
     # Timeunit is milliseconds.
     BACKOFF_INITIAL_DELAY = 1000
     MAX_BACKOFF_DELAY = 1024000
     logger = None
+    logging_handler = None
 
     def __init__(self, api_key, proxy=None, timeout=None, debug=False):
         """ api_key : google api key
@@ -210,10 +214,16 @@ class GCM(object):
         :return: the handler after adding it
         """
         if not handler:
-            handler = logging.StreamHandler()
-            handler.setFormatter(logging.Formatter('[%(asctime)s - %(levelname)s - %(filename)s:%(lineno)s - %(funcName)s()] %(message)s'))
+            # Use a singleton logging_handler instead of recreating it,
+            # so we can remove-and-re-add safely without having duplicate handlers
+            if GCM.logging_handler is None:
+                GCM.logging_handler = logging.StreamHandler()
+                GCM.logging_handler.setFormatter(logging.Formatter(
+                    '[%(asctime)s - %(levelname)s - %(filename)s:%(lineno)s - %(funcName)s()] %(message)s'))
+            handler = GCM.logging_handler
 
         GCM.logger = logging.getLogger(__name__)
+        GCM.logger.removeHandler(handler)
         GCM.logger.addHandler(handler)
         GCM.logger.setLevel(level)
         GCM.log('Added a stderr logging handler to logger: {0}', __name__)
@@ -221,16 +231,18 @@ class GCM(object):
         # Enable requests logging
         requests_logger_name = 'requests.packages.urllib3'
         requests_logger = logging.getLogger(requests_logger_name)
+        requests_logger.removeHandler(handler)
         requests_logger.addHandler(handler)
         requests_logger.setLevel(level)
         GCM.log('Added a stderr logging handler to logger: {0}', requests_logger_name)
 
     @staticmethod
     def log(message, *data):
-        if GCM.logger:
+        if GCM.logger and message:
             GCM.logger.debug(message.format(*data))
 
-    def construct_payload(self, **kwargs):
+    @staticmethod
+    def construct_payload(**kwargs):
         """
         Construct the dictionary mapping of parameters.
         Encodes the dictionary into JSON if for json requests.
@@ -244,25 +256,27 @@ class GCM(object):
         if is_json:
             if 'topic' not in kwargs and 'registration_ids' not in kwargs:
                 raise GCMMissingRegistrationException("Missing registration_ids or topic")
-            elif 'topic' in kwargs and 'registration_ids' in kwargs :
-                raise GCMInvalidInputException("Invalid parameters! Can't have both 'registration_ids' and 'to' as input parameters")
+            elif 'topic' in kwargs and 'registration_ids' in kwargs:
+                raise GCMInvalidInputException(
+                    "Invalid parameters! Can't have both 'registration_ids' and 'to' as input parameters")
 
             if 'topic' in kwargs:
                 kwargs['to'] = '/topics/{}'.format(kwargs.pop('topic'))
             elif 'registration_ids' not in kwargs:
-                    raise GCMMissingRegistrationException("Missing registration_ids")
-            
+                raise GCMMissingRegistrationException("Missing registration_ids")
+
             payload = JsonPayload(**kwargs).body
         else:
             payload = PlaintextPayload(**kwargs).body
 
         return payload
 
-    def make_request(self, data, is_json=True):
+    def make_request(self, data, is_json=True, session=None):
         """
         Makes a HTTP request to GCM servers with the constructed payload
 
         :param data: return value from construct_payload method
+        :param session: requests.Session object to use for request (optional)
         :raises GCMMalformedJsonException: if malformed JSON request found
         :raises GCMAuthenticationException: if there was a problem with authentication, invalid api key
         :raises GCMConnectionException: if GCM is screwed
@@ -284,10 +298,18 @@ class GCM(object):
         GCM.log('Request data: {0}', data)
         GCM.log('Request is_json: {0}', is_json)
 
-        response = requests.post(
-            self.url, data=data, headers=headers,
-            proxies=self.proxy, timeout=self.timeout,
-        )
+        new_session = None
+        if not session:
+            session = new_session = requests.Session()
+
+        try:
+            response = session.post(
+                    self.url, data=data, headers=headers,
+                    proxies=self.proxy, timeout=self.timeout,
+            )
+        finally:
+            if new_session:
+                new_session.close()
 
         GCM.log('Response status: {0} {1}', response.status_code, response.reason)
         GCM.log('Response headers: {0}', response.headers)
@@ -307,30 +329,31 @@ class GCM(object):
         # Failures
         if response.status_code == 400:
             raise GCMMalformedJsonException(
-                "The request could not be parsed as JSON")
+                    "The request could not be parsed as JSON")
         elif response.status_code == 401:
             raise GCMAuthenticationException(
-                "There was an error authenticating the sender account")
+                    "There was an error authenticating the sender account")
         elif response.status_code == 503:
             raise GCMUnavailableException("GCM service is unavailable")
         else:
             error = "GCM service error: %d" % response.status_code
             raise GCMUnavailableException(error)
 
-    def raise_error(self, error):
+    @staticmethod
+    def raise_error(error):
         if error == 'InvalidRegistration':
             raise GCMInvalidRegistrationException("Registration ID is invalid")
         elif error == 'Unavailable':
             # Plain-text requests will never return Unavailable as the error code.
             # http://developer.android.com/guide/google/gcm/gcm.html#error_codes
             raise GCMUnavailableException(
-                "Server unavailable. Resent the message")
+                    "Server unavailable. Resent the message")
         elif error == 'NotRegistered':
             raise GCMNotRegisteredException(
-                "Registration id is not valid anymore")
+                    "Registration id is not valid anymore")
         elif error == 'MismatchSenderId':
             raise GCMMismatchSenderIdException(
-                "A Registration ID is tied to a certain group of senders")
+                    "A Registration ID is tied to a certain group of senders")
         elif error == 'MessageTooBig':
             raise GCMMessageTooBigException("Message can't exceed 4096 bytes")
         elif error == 'MissingRegistration':
@@ -349,14 +372,18 @@ class GCM(object):
 
         # Split the first line by =
         key, value = response_lines[0].split('=')
+
+        # Error on first line
         if key == 'Error':
             self.raise_error(value)
-        else:
+        else:  # Canonical_id from the second line
             if len(response_lines) == 2:
-                return response_lines[1].split('=')[1]
-            return
+                return unquote(response_lines[1].split('=')[1])
+            return None  # TODO: Decide a way to return message id without breaking backwards compatibility
+            # unquote(value)  # ID of the sent message (from the first line)
 
-    def handle_json_response(self, response, registration_ids):
+    @staticmethod
+    def handle_json_response(response, registration_ids):
         errors = group_response(response, registration_ids, 'error')
         canonical = group_response(response, registration_ids, 'registration_id')
         success = group_response(response, registration_ids, 'message_id')
@@ -374,13 +401,15 @@ class GCM(object):
 
         return info
 
-    def handle_topic_response(self, response):
+    @staticmethod
+    def handle_topic_response(response):
         error = response.get('error')
         if error:
             raise GCMTopicMessageException(error)
         return response['message_id']
 
-    def extract_unsent_reg_ids(self, info):
+    @staticmethod
+    def extract_unsent_reg_ids(info):
         if 'errors' in info and 'Unavailable' in info['errors']:
             return info['errors']['Unavailable']
         return []
@@ -398,6 +427,7 @@ class GCM(object):
 
         kwargs['is_json'] = False
         retries = kwargs.pop('retries', 5)
+        session = kwargs.pop('session', None)
         payload = self.construct_payload(**kwargs)
         backoff = self.BACKOFF_INITIAL_DELAY
         info = None
@@ -405,23 +435,27 @@ class GCM(object):
 
         for attempt in range(retries):
             try:
-                response = self.make_request(payload, is_json=False)
+                response = self.make_request(payload, is_json=False, session=session)
                 info = self.handle_plaintext_response(response)
                 has_error = False
             except GCMUnavailableException:
                 has_error = True
 
             if self.retry_after:
-                GCM.log("[Attempt #{0}] Retry-After ~> Sleeping for {1} seconds".format(attempt, self.retry_after))
+                GCM.log("[plaintext_request - Attempt #{0}] Retry-After ~> Sleeping for {1} seconds".format(attempt,
+                                                                                                            self.retry_after))
                 time.sleep(self.retry_after)
                 self.retry_after = None
             elif has_error:
                 sleep_time = backoff / 2 + random.randrange(backoff)
                 nap_time = float(sleep_time) / 1000
-                GCM.log("[Attempt #{0}]Backoff ~> Sleeping for {1} seconds".format(attempt, nap_time))
+                GCM.log(
+                    "[plaintext_request - Attempt #{0}]Backoff ~> Sleeping for {1} seconds".format(attempt, nap_time))
                 time.sleep(nap_time)
                 if 2 * backoff < self.MAX_BACKOFF_DELAY:
                     backoff *= 2
+            else:
+                break
 
         if has_error:
             raise IOError("Could not make request after %d attempts" % retries)
@@ -443,6 +477,7 @@ class GCM(object):
         args = dict(**kwargs)
 
         retries = args.pop('retries', 5)
+        session = args.pop('session', None)
         payload = self.construct_payload(**args)
         registration_ids = args['registration_ids']
         backoff = self.BACKOFF_INITIAL_DELAY
@@ -451,7 +486,7 @@ class GCM(object):
 
         for attempt in range(retries):
             try:
-                response = self.make_request(payload, is_json=True)
+                response = self.make_request(payload, is_json=True, session=session)
                 info = self.handle_json_response(response, registration_ids)
                 unsent_reg_ids = self.extract_unsent_reg_ids(info)
                 has_error = False
@@ -467,13 +502,14 @@ class GCM(object):
                 payload = self.construct_payload(**args)
 
                 if self.retry_after:
-                    GCM.log("[Attempt #{0}] Retry-After ~> Sleeping for {1}".format(attempt, self.retry_after))
+                    GCM.log("[json_request - Attempt #{0}] Retry-After ~> Sleeping for {1}".format(attempt,
+                                                                                                   self.retry_after))
                     time.sleep(self.retry_after)
                     self.retry_after = None
                 else:
                     sleep_time = backoff / 2 + random.randrange(backoff)
                     nap_time = float(sleep_time) / 1000
-                    GCM.log("[Attempt #{0}] Backoff ~> Sleeping for {1}".format(attempt, nap_time))
+                    GCM.log("[json_request - Attempt #{0}] Backoff ~> Sleeping for {1}".format(attempt, nap_time))
                     time.sleep(nap_time)
                     if 2 * backoff < self.MAX_BACKOFF_DELAY:
                         backoff *= 2
@@ -485,6 +521,9 @@ class GCM(object):
 
         return info
 
+    def send_downstream_message(self, **kwargs):
+        return self.json_request(**kwargs)
+
     def send_topic_message(self, **kwargs):
         """
         Publish Topic Messaging to GCM servers
@@ -501,23 +540,27 @@ class GCM(object):
             raise GCMInvalidInputException("Topic name cannot be empty!")
 
         retries = kwargs.pop('retries', 5)
+        session = kwargs.pop('session', None)
         payload = self.construct_payload(**kwargs)
         backoff = self.BACKOFF_INITIAL_DELAY
 
         for attempt in range(retries):
             try:
-                response = self.make_request(payload, is_json=True)
+                response = self.make_request(payload, is_json=True, session=session)
                 return self.handle_topic_response(response)
-            except GCMUnavailableException:
-                sleep_time = backoff / 2 + random.randrange(backoff)
-                time.sleep(float(sleep_time) / 1000)
-                if 2 * backoff < self.MAX_BACKOFF_DELAY:
-                    backoff *= 2
-        else:
-            raise IOError("Could not make request after %d attempts" % retries)
+            except (GCMUnavailableException, GCMTopicMessageException):
+                if self.retry_after:
+                    GCM.log("[send_topic_message - Attempt #{0}] Retry-After ~> Sleeping for {1}"
+                            .format(attempt, self.retry_after))
 
-    def send_device_group_message(self, **kwargs):
-        raise NotImplementedError
+                    time.sleep(self.retry_after)
+                    self.retry_after = None
+                else:
+                    sleep_time = backoff / 2 + random.randrange(backoff)
+                    nap_time = float(sleep_time) / 1000
+                    GCM.log("[send_topic_message - Attempt #{0}] Backoff ~> Sleeping for {1}".format(attempt, nap_time))
+                    time.sleep(nap_time)
+                    if 2 * backoff < self.MAX_BACKOFF_DELAY:
+                        backoff *= 2
 
-    def send_downstream_message(self, **kwargs):
-        return self.json_request(**kwargs)
+        raise IOError("Could not make request after %d attempts" % retries)
